/*

		MultiMediaCard Driver Module
		for Fox 2.6 kernel

		(c) 2007 AFM Design - www.afmdesign.net

	License:
	    This program is free software: you can redistribute it and/or modify	
	    it under the terms of the GNU General Public License as published by
	    the Free Software Foundation, either version 3 of the License, or
	    (at your option) any later version.

	    This program is distributed in the hope that it will be useful,
	    but WITHOUT ANY WARRANTY; without even the implied warranty of
	    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	    GNU General Public License for more details.
	
	    You should have received a copy of the GNU General Public License
	    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/


// Includes

#include <linux/string.h>
#include <linux/kernel.h>

#include "mmc_def.h"
#include "mmc_ctrl.h"
#include "mmc_spi.h"


// Defines



// Globals

unsigned char mmc_block_mode;


// Functions

int mmc_extract_data(unsigned char *buf, int buf_bits, int start_bit, int bit_len)
{
	unsigned int data, m;
	int n, s, l;

	n = ((buf_bits - 1 - start_bit) / 8);
	s = 7 - ((buf_bits - 1 - start_bit) % 8);
	l = s + bit_len;
	m = 0;
	while (bit_len--) m = (m << 1) + 1;

	data = buf[n];
	if (l > 8) data |= ((unsigned int)buf[n - 1] << 8);
	if (l > 16) data |= ((unsigned int)buf[n - 2] << 16);
	if (l > 24) data |= ((unsigned int)buf[n - 3] << 24);
	data >>= s;
	data &= m;

	return (int)data;
}


int mmc_init_card(void)
{
	int i, r1;
	unsigned char v2_card;
	unsigned char hcsd;
	unsigned char csd_version;
	unsigned char if_cond_buf[MMC_IF_COND_SIZE];
	unsigned char ocr_buf[MMC_OCR_SIZE];
	unsigned char cid_buf[MMC_CID_SIZE];
	unsigned char csd_buf[MMC_CSD_SIZE];
	unsigned char cid_pnm[6];
	unsigned int size;
	unsigned int capacity;

	mmc_spi_rate(MMC_SPI_SLOW);
	mmc_spi_clk(10);								// At least 74 clk's to start communication

	r1 = mmc_cmd(MMC_GO_IDLE_STATE, 0); 
	if (r1 == -1) return -1;
	if ((r1 & MMC_R1_IN_IDLE_STATE) != MMC_R1_IN_IDLE_STATE) return -1;

	r1 = mmc_cmd_read(MMC_SEND_IF_COND, MMC_IF_COND_MAGIC, if_cond_buf, sizeof(if_cond_buf)); 
	if (r1 == -1) return -1;
	if ((r1 & MMC_R1_ILLEGAL_COMMAND) == MMC_R1_ILLEGAL_COMMAND) v2_card = 0;
	else if (mmc_extract_data(if_cond_buf, (sizeof(if_cond_buf) * 8), 0, 12) == MMC_IF_COND_MAGIC) {
		v2_card = 1;
	} else {
		printk(KERN_WARNING "mmc: Cannot determine card type\n");
		return -1;
	}

	r1 = mmc_cmd_read(MMC_READ_OCR, 0, ocr_buf, sizeof(ocr_buf)); 
	if (r1 == -1) return -1;
	if (mmc_extract_data(ocr_buf, (MMC_OCR_SIZE * 8), 20, 2) == 0) {
		printk(KERN_WARNING "mmc: Card voltage range incompatible\n");
		return -1;
	}

	i = 0;
	do {
		if (v2_card) {								// SD / HCSD Card
			r1 = mmc_cmd(MMC_APP_CMD, 0); 
			if (r1 == -1) return -1;
			r1 = mmc_cmd(MMC_SD_SEND_OP_COND, MMC_OP_COND_HCS); 
			if (r1 == -1) return -1;
		} else {									// MMC
			r1 = mmc_cmd(MMC_SEND_OP_COND, 0); 
			if (r1 == -1) return -1;
		}
		if (i++ == MMC_MAX_OP_COND_LOOPS) {
			printk(KERN_WARNING "mmc: op_cond_loops exceeded\n");
			return -1;
		}
	} while ((r1 & MMC_R1_IN_IDLE_STATE) == MMC_R1_IN_IDLE_STATE);

	mmc_spi_rate(MMC_SPI_FAST);

	hcsd = 0;
	if (v2_card) {
		r1 = mmc_cmd_read(MMC_READ_OCR, 0, ocr_buf, sizeof(ocr_buf)); 
		if (r1 == -1) return -1;
		if (mmc_extract_data(ocr_buf, (MMC_OCR_SIZE * 8), 30, 1) == 1) hcsd = 1;
	}

	r1 = mmc_cmd_read_reg(MMC_SEND_CID, 0, cid_buf, sizeof(cid_buf)); 
	if (r1 == -1) return -1;

	r1 = mmc_cmd_read_reg(MMC_SEND_CSD, 0, csd_buf, sizeof(csd_buf)); 
	if (r1 == -1) return -1;

	csd_version = mmc_extract_data(csd_buf, (MMC_CSD_SIZE * 8), 126, 2);
	if (csd_version == 0) {				// CSD Version 1.0
		size = 	(1 << mmc_extract_data(csd_buf, 128, 80, 4)) *		// read_bl_len
				(mmc_extract_data(csd_buf, 128, 62, 12) + 1) *			// c_size
				(1 << (mmc_extract_data(csd_buf, 128, 47, 3) + 2));		// c_size_mult
		mmc_block_mode = 0;
	} else if (csd_version == 1) {		// CSD Version 2.0
		size = (mmc_extract_data(csd_buf, 128, 48, 22) + 1)  * (512 * 1024);
		mmc_block_mode = 1;
	} else {
		printk(KERN_WARNING "mmc: csd_version incompatible\n");
		return -1;
	}
	capacity = size / MMC_KERNEL_SECTORSIZE;

	memcpy(cid_pnm, cid_buf + 3, 5);
	cid_pnm[5] = '\0';

	printk(KERN_INFO "mmc: %s, size %u bytes (%u MiB) %s\n", cid_pnm,	size, size / (1024 * 1024),
		(hcsd == 1) ? "(HCSD)" : ((v2_card == 1) ? "(SD)" : "(MMC)"));
	printk(KERN_INFO "mmc: Serial# %u, MID = %d, '%c%c'\n", mmc_extract_data(cid_buf, (MMC_CID_SIZE * 8), 16, 32),
 		mmc_extract_data(cid_buf, (MMC_CID_SIZE * 8), 120, 8), cid_buf[1], cid_buf[2]);

	return capacity;
}

int mmc_read_sectors(unsigned char *buf, unsigned int start, unsigned int num)
{
	unsigned int addr;
	unsigned int addrstep = 1;
	unsigned int cnt = 0;

//printk("Sector read (%d, %d)\n", start, num);

	if (mmc_block_mode == 0) {
		addr = start * MMC_KERNEL_SECTORSIZE;
		addrstep = MMC_KERNEL_SECTORSIZE;
	} else addr = start;

	while (num--) {
		if ((mmc_read_block(buf, addr)) == -1) {
			mmc_idle();
			break;
		}
		buf += MMC_KERNEL_SECTORSIZE;
		addr += addrstep;
		cnt +=  1;
	}

	return cnt;
}


int mmc_write_sectors(unsigned char *buf, unsigned int start, unsigned int num)
{
	unsigned int addr;
	unsigned int addrstep = 1;
	unsigned int cnt = 0;

//printk("Sector write (%d, %d)\n", start, num);

	if (mmc_block_mode == 0) {
		addr = start * MMC_KERNEL_SECTORSIZE;
		addrstep = MMC_KERNEL_SECTORSIZE;
	} else addr = start;

	while (num--) {
		if ((mmc_write_block(buf, addr)) == -1) {
			mmc_idle();
			break;
		}
		buf += MMC_KERNEL_SECTORSIZE;
		addr += addrstep;
		cnt +=  1;
	}

	return cnt;
}
