/*

		MultiMediaCard Driver Module
		for Fox 2.6 kernel

		Low level software SPI functions


		(c) 2007 AFM Design - www.afmdesign.net

	License:
	    This program is free software: you can redistribute it and/or modify	
	    it under the terms of the GNU General Public License as published by
	    the Free Software Foundation, either version 3 of the License, or
	    (at your option) any later version.

	    This program is distributed in the hope that it will be useful,
	    but WITHOUT ANY WARRANTY; without even the implied warranty of
	    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	    GNU General Public License for more details.
	
	    You should have received a copy of the GNU General Public License
	    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

// Includes

#include <linux/kernel.h>
#include <asm/delay.h>
#include <asm/io.h>
#include <asm/arch/io_interface_mux.h>

#include "mmc_spi.h"

// Typedefs

unsigned char mmc_spi_read_byte_slow(void);
void mmc_spi_write_byte_slow(unsigned char c);


// Globals

static unsigned char (*mmc_spi_read_func)(void) = mmc_spi_read_byte_slow;
static void (*mmc_spi_write_func)(unsigned char) = mmc_spi_write_byte_slow;
static unsigned char mmc_led;


// Functions

void mmc_led_on(void)
{
	 MMC_LED_ON;
}

void mmc_led_off(void) 
{
	MMC_LED_OFF;
}

void mmc_spi_setup(void)
{
	MMC_DESELECT;
	MMC_CLK_HIGH;
	MMC_SO_HIGH;
	MMC_LED_SETUP;
	MMC_LED_OFF;
}

void mmc_spi_write_byte_slow(unsigned char c)
{
	unsigned char m;

	for (m = 0x80; (m); m >>= 1) {
		MMC_CLK_LOW;
		udelay(1);
		if (c & m) MMC_SO_HIGH;
		else MMC_SO_LOW;
		MMC_CLK_HIGH;
		udelay(1);
	}
}

void mmc_spi_write_byte(unsigned char c)
{
	unsigned char m;

	for (m = 0x80; (m); m >>= 1) {
		MMC_CLK_LOW;
		if (c & m) MMC_SO_HIGH;
		else MMC_SO_LOW;
		MMC_CLK_HIGH;
	}
}

void mmc_spi_write_n(void *buffer, unsigned short len)
{
	MMC_LED_TOGGLE;
	while (len--) (*mmc_spi_write_func)(*(unsigned char *)buffer++);
	MMC_SO_HIGH;
	MMC_LED_TOGGLE;
}

void mmc_spi_write(unsigned char c)
{
	mmc_spi_write_n(&c, 1);
}

unsigned char mmc_spi_read_byte_slow(void)
{
	unsigned char m, d;

	for (m = 0x80, d = 0; (m); m >>= 1) {
		MMC_CLK_LOW;
		udelay(1);
		MMC_CLK_HIGH;
		udelay(1);
		if (MMC_SI_PIN) d |= m;
	}
	return d;
}

unsigned char mmc_spi_read_byte(void)
{
	unsigned char m, d;

	for (m = 0x80, d = 0; (m); m >>= 1) {
		MMC_CLK_LOW;
		MMC_CLK_HIGH;
		if (MMC_SI_PIN) d |= m;
	}
	return d;
}

void mmc_spi_read_n(void *buffer, unsigned short len)
{
	MMC_LED_TOGGLE;
	while (len--) {
		*(unsigned char *)buffer++ = (*mmc_spi_read_func)();
	}
	MMC_LED_TOGGLE;
}

unsigned char mmc_spi_read(void)
{
	unsigned char c;

	mmc_spi_read_n(&c, 1);
	return c;
}

void mmc_spi_clk(unsigned short n)
{
	while (n--) mmc_spi_read();
}

void mmc_spi_rate(unsigned short rate)
{
	switch (rate) {
		case MMC_SPI_SLOW:
			mmc_spi_read_func = mmc_spi_read_byte_slow;
			mmc_spi_write_func = mmc_spi_write_byte_slow;
			break;

		case MMC_SPI_FAST:
			mmc_spi_read_func = mmc_spi_read_byte;
			mmc_spi_write_func = mmc_spi_write_byte;
			break;

		default:
			printk(KERN_WARNING "mmc: Invalid rate selected\n");
			break;
	}
}

